@@ -46,6 +46,8 @@
 
 #include "cairo-clip-private.h"
 #include "cairo-device-private.h"
+#include "cairo-default-context-private.h"
+#include "cairo-image-surface-private.h"
 #include "cairo-error-private.h"
 #include "cairo-output-stream-private.h"
 #include "cairo-recording-surface-private.h"
@@ -156,10 +158,13 @@ static const char *
 _antialias_to_string (cairo_antialias_t antialias)
 {
     static const char *names[] = {
-	"ANTIALIAS_DEFAULT",	/* CAIRO_ANTIALIAS_DEFAULT */
-	"ANTIALIAS_NONE",	/* CAIRO_ANTIALIAS_NONE */
-	"ANTIALIAS_GRAY",	/* CAIRO_ANTIALIAS_GRAY */
-	"ANTIALIAS_SUBPIXEL"	/* CAIRO_ANTIALIAS_SUBPIXEL */
+	"DEFAULT",	/* CAIRO_ANTIALIAS_DEFAULT */
+	"NONE",         /* CAIRO_ANTIALIAS_NONE */
+	"GRAY",         /* CAIRO_ANTIALIAS_GRAY */
+	"SUBPIXEL",	/* CAIRO_ANTIALIAS_SUBPIXEL */
+	"FAST",         /* CAIRO_ANTIALIAS_FAST */
+	"GOOD",         /* CAIRO_ANTIALIAS_GOOD */
+	"BEST",         /* CAIRO_ANTIALIAS_BEST */
     };
     assert (antialias < ARRAY_LENGTH (names));
     return names[antialias];
@@ -205,6 +210,7 @@ _format_to_string (cairo_format_t format)
 {
     switch (format) {
     case CAIRO_FORMAT_ARGB32:  return "ARGB32";
+    case CAIRO_FORMAT_RGB30:   return "RGB30";
     case CAIRO_FORMAT_RGB24:   return "RGB24";
     case CAIRO_FORMAT_RGB16_565:   return "RGB16_565";
     case CAIRO_FORMAT_A8:      return "A8";
@@ -415,20 +421,19 @@ _cairo_xml_close_path (void *closure)
 
 static void
 _cairo_xml_emit_path (cairo_xml_t *xml,
-		      cairo_path_fixed_t *path)
+		      const cairo_path_fixed_t *path)
 {
     cairo_status_t status;
 
     _cairo_xml_printf_start (xml, "<path>");
     status = _cairo_path_fixed_interpret (path,
-					CAIRO_DIRECTION_FORWARD,
 					_cairo_xml_move_to,
 					_cairo_xml_line_to,
 					_cairo_xml_curve_to,
 					_cairo_xml_close_path,
 					xml);
     assert (status == CAIRO_STATUS_SUCCESS);
-    _cairo_xml_printf_start (xml, "</path>");
+    _cairo_xml_printf_end (xml, "</path>");
 }
 
 static void
@@ -500,7 +505,7 @@ _cairo_xml_surface_emit_clip_path (cairo_xml_surface_t *surface,
 
 static cairo_status_t
 _cairo_xml_surface_emit_clip (cairo_xml_surface_t *surface,
-			      cairo_clip_t *clip)
+			      const cairo_clip_t *clip)
 {
     if (clip == NULL)
 	return CAIRO_STATUS_SUCCESS;
@@ -555,10 +560,8 @@ _cairo_xml_emit_linear (cairo_xml_t *xml,
 {
     _cairo_xml_printf (xml,
 		       "<linear x1='%f' y1='%f' x2='%f' y2='%f'>",
-		       _cairo_fixed_to_double (linear->p1.x),
-		       _cairo_fixed_to_double (linear->p1.y),
-		       _cairo_fixed_to_double (linear->p2.x),
-		       _cairo_fixed_to_double (linear->p2.y));
+		       linear->pd1.x, linear->pd1.y,
+		       linear->pd2.x, linear->pd2.y);
     _cairo_xml_indent (xml, 2);
     _cairo_xml_emit_gradient (xml, &linear->base);
     _cairo_xml_indent (xml, -2);
@@ -572,12 +575,8 @@ _cairo_xml_emit_radial (cairo_xml_t *xml,
 {
     _cairo_xml_printf (xml,
 		       "<radial x1='%f' y1='%f' r1='%f' x2='%f' y2='%f' r2='%f'>",
-		       _cairo_fixed_to_double (radial->c1.x),
-		       _cairo_fixed_to_double (radial->c1.y),
-		       _cairo_fixed_to_double (radial->r1),
-		       _cairo_fixed_to_double (radial->c2.x),
-		       _cairo_fixed_to_double (radial->c2.y),
-		       _cairo_fixed_to_double (radial->r2));
+		       radial->cd1.center.x, radial->cd1.center.y, radial->cd1.radius,
+		       radial->cd2.center.x, radial->cd2.center.y, radial->cd2.radius);
     _cairo_xml_indent (xml, 2);
     _cairo_xml_emit_gradient (xml, &radial->base);
     _cairo_xml_indent (xml, -2);
@@ -692,7 +691,7 @@ static cairo_int_status_t
 _cairo_xml_surface_paint (void			*abstract_surface,
 			  cairo_operator_t	 op,
 			  const cairo_pattern_t	*source,
-			  cairo_clip_t		*clip)
+			  const cairo_clip_t	*clip)
 {
     cairo_xml_surface_t *surface = abstract_surface;
     cairo_xml_t *xml = to_xml (surface);
@@ -719,10 +718,10 @@ _cairo_xml_surface_paint (void			*abstract_surface,
 
 static cairo_int_status_t
 _cairo_xml_surface_mask (void			*abstract_surface,
-			  cairo_operator_t	 op,
-			  const cairo_pattern_t	*source,
-			  const cairo_pattern_t	*mask,
-			  cairo_clip_t		*clip)
+			 cairo_operator_t	 op,
+			 const cairo_pattern_t	*source,
+			 const cairo_pattern_t	*mask,
+			 const cairo_clip_t	*clip)
 {
     cairo_xml_surface_t *surface = abstract_surface;
     cairo_xml_t *xml = to_xml (surface);
@@ -755,13 +754,13 @@ static cairo_int_status_t
 _cairo_xml_surface_stroke (void				*abstract_surface,
 			   cairo_operator_t		 op,
 			   const cairo_pattern_t	*source,
-			   cairo_path_fixed_t		*path,
+			   const cairo_path_fixed_t		*path,
 			   const cairo_stroke_style_t		*style,
 			   const cairo_matrix_t		*ctm,
 			   const cairo_matrix_t		*ctm_inverse,
 			   double			 tolerance,
 			   cairo_antialias_t		 antialias,
-			   cairo_clip_t			*clip)
+			   const cairo_clip_t		*clip)
 {
     cairo_xml_surface_t *surface = abstract_surface;
     cairo_xml_t *xml = to_xml (surface);
@@ -811,11 +810,11 @@ static cairo_int_status_t
 _cairo_xml_surface_fill (void			*abstract_surface,
 			 cairo_operator_t	 op,
 			 const cairo_pattern_t	*source,
-			 cairo_path_fixed_t	*path,
+			 const cairo_path_fixed_t*path,
 			 cairo_fill_rule_t	 fill_rule,
 			 double			 tolerance,
 			 cairo_antialias_t	 antialias,
-			 cairo_clip_t		*clip)
+			 const cairo_clip_t	*clip)
 {
     cairo_xml_surface_t *surface = abstract_surface;
     cairo_xml_t *xml = to_xml (surface);
@@ -872,7 +871,7 @@ _cairo_xml_emit_type42_font (cairo_xml_t *xml,
     if (unlikely (buf == NULL))
 	return _cairo_error (CAIRO_STATUS_NO_MEMORY);
 
-    status = backend->load_truetype_table (scaled_font, 0, 0, buf, NULL);
+    status = backend->load_truetype_table (scaled_font, 0, 0, buf, &size);
     if (unlikely (status)) {
 	free (buf);
 	return status;
@@ -930,7 +929,7 @@ _cairo_xml_emit_scaled_font (cairo_xml_t *xml,
 			     cairo_glyph_t *glyphs,
 			     int num_glyphs)
 {
-    cairo_status_t status;
+    cairo_int_status_t status;
 
     _cairo_xml_printf (xml, "<scaled-font>");
     _cairo_xml_indent (xml, 2);
@@ -954,8 +953,7 @@ _cairo_xml_surface_glyphs (void			    *abstract_surface,
 			   cairo_glyph_t	    *glyphs,
 			   int			     num_glyphs,
 			   cairo_scaled_font_t	    *scaled_font,
-			   cairo_clip_t		    *clip,
-			   int			    *remaining_glyphs)
+			   const cairo_clip_t       *clip)
 {
     cairo_xml_surface_t *surface = abstract_surface;
     cairo_xml_t *xml = to_xml (surface);
@@ -989,48 +987,41 @@ _cairo_xml_surface_glyphs (void			    *abstract_surface,
     _cairo_xml_indent (xml, -2);
     _cairo_xml_printf (xml, "</glyphs>");
 
-    *remaining_glyphs = 0;
     return CAIRO_STATUS_SUCCESS;
 }
 
 static const cairo_surface_backend_t
 _cairo_xml_surface_backend = {
     CAIRO_SURFACE_TYPE_XML,
-    _cairo_xml_surface_create_similar,
     NULL,
-    NULL, NULL, /* source image */
-    NULL, NULL, /* dst image */
-    NULL, /* clone_similar */
-    NULL, /* composite */
-    NULL, /* fill_rectangles */
-    NULL, /* composite_trapezoids */
-    NULL, /* create_span_renderer */
-    NULL, /* check_span_renderer */
-    NULL, NULL, /* copy/show page */
+
+    _cairo_default_context_create,
+
+    _cairo_xml_surface_create_similar,
+    NULL, /* create_similar_image */
+    NULL, /* map_to_image */
+    NULL, /* unmap_image */
+
+    _cairo_surface_default_source,
+    NULL, /* acquire source image */
+    NULL, /* release source image */
+    NULL, /* snapshot */
+
+    NULL, /* copy page */
+    NULL, /* show page */
+
     _cairo_xml_surface_get_extents,
-    NULL, /* old_show_glyphs */
     NULL, /* get_font_options */
+
     NULL, /* flush */
     NULL, /* mark_dirty_rectangle */
-    NULL, /* font fini */
-    NULL, /* scaled_glyph_fini */
 
-    /* The 5 high level operations */
     _cairo_xml_surface_paint,
     _cairo_xml_surface_mask,
     _cairo_xml_surface_stroke,
     _cairo_xml_surface_fill,
-    _cairo_xml_surface_glyphs,
-
-    NULL, /* snapshot */
-
-    NULL, /* is_similar */
     NULL, /* fill_stroke */
-    NULL, /* create_solid_pattern_surface */
-    NULL, /* can_repaint_solid_pattern_surface */
-
-    /* The alternate high-level text operation */
-    NULL, NULL, /* has, show_text_glyphs */
+    _cairo_xml_surface_glyphs,
 };
 
 static cairo_surface_t *
